#![no_std]
#![no_main]

use ch32_hal::{self as hal, bind_interrupts};
use core::borrow::{Borrow, BorrowMut};
use core::cell::{Cell, RefCell};
use core::panic::PanicInfo;
use critical_section::Mutex;
use embassy_executor::Spawner;
use embassy_futures::join::join;
use embassy_time::{Duration, Ticker, Timer};
use embassy_usb::class::cdc_acm::{CdcAcmClass, State};
use embassy_usb::driver::EndpointError;
use embassy_usb::Builder;
use hal::gpio::{AnyPin, Level, Output, Pin};
use hal::pac;
use hal::pac::gpio::regs::Outdr;
use hal::pac::gpio::{self, Gpio};
use hal::println;
use hal::time::Hertz;
use hal::timer::low_level::CountingMode;
use hal::timer::simple_pwm::{PwmPin, SimplePwm};
use hal::{peripherals, usbd};
use heapless::{binary_heap::Min, BinaryHeap};
use micromath::F32Ext;
use portable_atomic::AtomicU64;

bind_interrupts!(struct Irqs {
    USB_LP_CAN1_RX0 => usbd::InterruptHandler<peripherals::USBD>;
});

const FPS: usize = 120;
const WIDTH: usize = 16;
const HEIGHT: usize = 16;
const EMPTY_FRAME: Frame = Frame::empty();

static CUR_FRAME: AtomicU64 = AtomicU64::new(0);
static MBI_QUEUE: Mutex<RefCell<BinaryHeap<MbiCommand, Min, 30>>> =
    Mutex::new(RefCell::new(BinaryHeap::new()));

/// Unfortunately embedded_hal doesn't support atomic writes to multiple bytes, which is required for this implementation, so we must take a platform register :(
struct MBI5043<SDI, DCLK, LE> {
    sdi: SDI,
    dclk: DCLK,
    le: LE,
    port: Gpio,
}

// Should be an enum containing the data itself, for auto generic
#[repr(u8)]
enum DataType {
    Data = 1,
    Global = 3,
    ReadConfig = 5,
    WriteConfig = 11,
    EnableWriteConfig = 15,
}

pub trait PinBit {
    fn bit(&self) -> u32;
}

impl<T: Pin> PinBit for T {
    fn bit(&self) -> u32 {
        1 << self.pin()
    }
}

impl<'a, SDI: Pin, DCLK: Pin, LE: Pin> MBI5043<SDI, DCLK, LE> {
    pub fn new(sdi: SDI, dclk: DCLK, le: LE) -> Self {
        assert_eq!(sdi.port(), dclk.port());
        assert_eq!(sdi.port(), le.port());

        let port = pac::GPIO(sdi.port() as usize);
        Self::set_port_mode(&sdi);
        Self::set_port_mode(&dclk);
        Self::set_port_mode(&le);
        Self {
            port,
            sdi,
            dclk,
            le,
        }
    }
    fn set_port_mode<T: Pin>(pin: &T) {
        if pin.pin() < 8 {
            let reg = pac::GPIO(pin.port() as usize).cfglr();
            // low bits
            reg.modify(|w| {
                w.set_mode(pin.pin() as usize, gpio::vals::Mode::OUTPUT_50MHZ);
                w.set_cnf(
                    pin.pin() as usize,
                    gpio::vals::Cnf::ANALOG_IN__PUSH_PULL_OUT,
                );
            });
        } else {
            // high bits
            let reg = pac::GPIO(pin.port() as usize).cfghr();
            reg.modify(|r| {
                r.set_mode(pin.pin() as usize, gpio::vals::Mode::OUTPUT_50MHZ);
                r.set_cnf(
                    pin.pin() as usize,
                    gpio::vals::Cnf::ANALOG_IN__PUSH_PULL_OUT,
                );
            });
        }
    }
    fn clk_data_le_mask(&self) -> u32 {
        self.dclk.bit() | self.sdi.bit() | self.le.bit()
    }
    fn set_all(&self) {
        self.port.outdr().modify(|p| {
            p.set_odr(self.dclk.pin() as usize, true);
            p.set_odr(self.sdi.pin() as usize, true);
            p.set_odr(self.le.pin() as usize, true)
        });
    }
    fn send_word<const LE_LEN: u8>(&self, val: &u16) {
        for i in (0..16).rev() {
            // clk low & new data
            self.port.outdr().modify(|w| {
                w.set_odr(self.dclk.pin() as usize, false);
                w.set_odr(self.sdi.pin() as usize, val & (1 << i) != 0);
                w.set_odr(self.le.pin() as usize, i < LE_LEN);
            });

            // rising clock
            self.port
                .outdr()
                .modify(|w| w.set_odr(self.dclk.pin() as usize, true));
        }
        // clear all bits
        self.port
            .outdr()
            .modify(|w| w.0 &= !self.clk_data_le_mask());
    }
}

pub trait MBI {
    fn send_pixels<T: Fn(u8) -> u16>(&self, pixels: &[u8], transform: T);
    fn set_gain(&self, gain: u8);
}

impl<SDI: Pin, DCLK: Pin, LE: Pin> MBI for MBI5043<SDI, DCLK, LE> {
    fn send_pixels<T: Fn(u8) -> u16>(&self, pixels: &[u8], transform: T) {
        for pixel in pixels {
            self.send_word::<{ DataType::Data as u8 }>(&transform(*pixel));
        }
        // dummy word to latch the output registers
        self.send_word::<{ DataType::Global as u8 }>(&0x0000);
    }
    // TODO: Need to store rest of the config and restore it here
    fn set_gain(&self, gain: u8) {
        let config = (gain as u16) << 4;
        self.send_word::<15>(&config);
        self.send_word::<11>(&config);
    }
}

#[derive(Debug)]
enum ProtocolState {
    WaitCmd,
    WaitLengthLsb,
    WaitLengthMsb,
    WaitData,
}

#[repr(u8)]
#[derive(Eq, PartialEq, Debug)]
enum Command {
    Info = 0,
    Frame = 1,
    Brightness = 2,
    Invalid = 255,
}

impl From<u8> for Command {
    fn from(value: u8) -> Self {
        match value {
            0 => Command::Info,
            1 => Command::Frame,
            2 => Command::Brightness,
            _ => Command::Invalid,
        }
    }
}

struct Protocol<'d, T: usbd::Instance + 'd> {
    class: &'d mut CdcAcmClass<'d, usbd::Driver<'d, T>>,
    state: ProtocolState,
    cmd: Command,
    to_read: u16,
    cur_pos: usize,
    buf: [u8; 1024],
}

impl<'d, T: usbd::Instance + 'd> Protocol<'d, T> {
    fn new(class: &'d mut CdcAcmClass<'d, usbd::Driver<'d, T>>) -> Self {
        Self {
            class,
            state: ProtocolState::WaitCmd,
            cmd: Command::Info,
            to_read: 0,
            cur_pos: 0,
            buf: [0; 1024],
        }
    }
    fn transition(&mut self, new_state: ProtocolState) {
        // println!("transition {:?} -> {:?}", self.state, new_state);
        self.state = new_state
    }
    async fn receive(&mut self) -> Result<(), Disconnected> {
        let mut buf = [0; 64];
        loop {
            let n = self.class.read_packet(&mut buf).await?;
            let mut data = &buf[..n];
            while !data.is_empty() {
                match self.state {
                    ProtocolState::WaitCmd => {
                        let cmd: Command = data[0].into();
                        if cmd != Command::Invalid {
                            self.cmd = cmd;
                            self.transition(ProtocolState::WaitLengthMsb);
                        }
                    }
                    ProtocolState::WaitLengthMsb => {
                        self.to_read = (data[0] as u16) << 8;
                        self.transition(ProtocolState::WaitLengthLsb);
                    }
                    ProtocolState::WaitLengthLsb => {
                        self.cur_pos = 0;
                        self.to_read |= data[0] as u16;
                        // println!("expecting {} bytes", self.to_read);
                        if self.to_read != 0 {
                            self.transition(ProtocolState::WaitData);
                        } else {
                            self.handle().await;
                            self.transition(ProtocolState::WaitCmd);
                        }
                    }
                    ProtocolState::WaitData => {
                        if self.to_read != 0 {
                            self.buf[self.cur_pos] = data[0];
                            self.cur_pos += 1;
                            self.to_read -= 1;
                        }
                        if self.to_read == 0 {
                            self.handle().await;
                            self.transition(ProtocolState::WaitCmd);
                        }
                    }
                }
                data = &data[1..];
            }
        }
    }
    async fn send_frame_no(&mut self) {
        let frame_be = CUR_FRAME
            .load(core::sync::atomic::Ordering::Relaxed)
            .to_be_bytes();
        self.class.write_packet(&frame_be).await.unwrap();
    }

    async fn handle(&mut self) {
        // println!("handling cmd: {:?}", self.cmd);
        match self.cmd {
            Command::Info => self.handle_info().await,
            Command::Frame => self.handle_frame().await,
            Command::Brightness => self.handle_brightness().await,
            Command::Invalid => unimplemented!(),
        }
    }

    // response:
    //   u16 width
    //   u16 height
    //   u16 fps
    //   u64 cur_frame
    async fn handle_info(&mut self) {
        let width_be = (WIDTH as u16).to_be_bytes();
        let height_be = (HEIGHT as u16).to_be_bytes();
        let fps_be = (FPS as u16).to_be_bytes();
        self.class.write_packet(&width_be).await.unwrap();
        self.class.write_packet(&height_be).await.unwrap();
        self.class.write_packet(&fps_be).await.unwrap();
        self.send_frame_no().await;
    }

    // data:
    //   u64 when
    //   u8[WIDTH * HEIGHT] pixels
    // response:
    //   u64 cur_frame
    async fn handle_frame(&mut self) {
        let cmd = MbiCommand::frame_from_packet(self.buf.as_slice());
        critical_section::with(|cs| MBI_QUEUE.borrow_ref_mut(cs).push(cmd));
        self.send_frame_no().await;
    }

    // data:
    //  u8 gain - 6 bits, 12.5 - 200%
    // response:
    //  u8 - 255 if ok, 0 if not
    async fn handle_brightness(&mut self) {
        let gain = self.buf[0];
        if gain & 0b11000000 != 0 {
            self.class.write_packet(&[0u8]).await.unwrap();
        } else {
            critical_section::with(|cs| {
                MBI_QUEUE
                    .borrow_ref_mut(cs)
                    .push(MbiCommand::from_brightness(gain))
            }).unwrap();
            self.class.write_packet(&[255u8]).await.unwrap();
        }
    }
}

#[embassy_executor::task]
async fn blink(pin1: AnyPin, pin2: AnyPin) {
    let mut leds = [
        Output::new(pin1, Level::Low, Default::default()),
        Output::new(pin2, Level::Low, Default::default()),
    ];
    let mut ticker = Ticker::every(Duration::from_millis(250));
    // println!("blinker running");
    let mut i = 0;
    loop {
        for idx in 0..leds.len() {
            leds[idx].set_level(((1 << idx) & i != 0).into());
        }
        i += 1;
        ticker.next().await
    }
}

// #[embassy_executor::task]
// async fn busy() {
//     loop {
//         pac::GPIOC.outdr().modify(|w| w.set_odr(13, !w.odr(13)));
//     }
// }

#[embassy_executor::task]
async fn usb(usb: peripherals::USBD, irq: Irqs, dp: peripherals::PA12, dm: peripherals::PA11) {
    let usb_driver = usbd::Driver::new(usb, irq, dp, dm);
    let usb_config = embassy_usb::Config::new(0xc0de, 0xcafe);

    let mut config_desc = [0; 256];
    let mut bos_desc = [0; 256];
    let mut control_buf = [0; 7];

    let mut state = State::new();
    let mut builder = Builder::new(
        usb_driver,
        usb_config,
        &mut config_desc,
        &mut bos_desc,
        &mut [],
        &mut control_buf,
    );

    let mut class = CdcAcmClass::new(&mut builder, &mut state, 64);
    let mut usb = builder.build();

    let usb_fut = usb.run();

    let echo_fut = async {
        let mut protocol = Protocol::new(&mut class);
        loop {
            protocol.class.wait_connection().await;
            protocol.receive().await.unwrap();
        }
    };
    join(usb_fut, echo_fut).await;
}

#[derive(Debug)]
struct Disconnected {}

impl From<EndpointError> for Disconnected {
    fn from(val: EndpointError) -> Self {
        match val {
            EndpointError::BufferOverflow => panic!("Buffer overflow"),
            EndpointError::Disabled => Disconnected {},
        }
    }
}

#[derive(Debug)]
struct MbiCommand {
    when: u64,
    inner: MbiCommandInner,
}

impl MbiCommand {
    fn frame_from_packet(packet: &[u8]) -> MbiCommand {
        Self {
            when: u64::from_be_bytes(packet[0..8].try_into().unwrap()),
            inner: Frame::from(&packet[8..]).into(),
        }
    }
    fn from_brightness(gain: u8) -> MbiCommand {
        Self {
            when: 0,
            inner: MbiCommandInner::Brightness(gain),
        }
    }
}

impl PartialOrd for MbiCommand {
    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
        self.when.partial_cmp(&other.when)
    }
}

impl PartialEq for MbiCommand {
    fn eq(&self, other: &Self) -> bool {
        self.when.eq(&other.when)
    }
}
impl Eq for MbiCommand {}
impl Ord for MbiCommand {
    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
        self.when.cmp(&other.when)
    }
}

#[derive(Debug)]
enum MbiCommandInner {
    Frame(Frame),
    Brightness(u8),
}

impl From<Frame> for MbiCommandInner {
    fn from(value: Frame) -> Self {
        MbiCommandInner::Frame(value)
    }
}

#[derive(Clone, Debug)]
struct Frame {
    pixels: [u8; WIDTH as usize * HEIGHT as usize],
}

impl Frame {
    const fn empty() -> Self {
        Self {
            pixels: [0; WIDTH as usize * HEIGHT as usize],
        }
    }
}

impl Default for Frame {
    fn default() -> Self {
        Self::empty()
    }
}

impl From<&[u8]> for Frame {
    fn from(value: &[u8]) -> Self {
        let mut pixels = [0; WIDTH as usize * HEIGHT as usize];
        for i in 0..pixels.len() {
            let y = i / WIDTH as usize;
            let x = WIDTH as usize - (i % WIDTH as usize) - 1;
            pixels[y * WIDTH as usize + x] = value[i];
        }
        Frame { pixels }
    }
}

struct GammaLut {
    gamma: f32,
    max: u16,
    lut: [u16; 256],
}

impl GammaLut {
    fn new(gamma: f32, max: u16) -> Self {
        let mut lut = [0u16; 256];
        let f32_max = max as f32;
        for i in 0..256 {
            let f_val = (i as f32 / 255.).powf(gamma) * f32_max + 0.5;
            lut[i] = f_val as u16;
        }
        Self { gamma, max, lut }
    }
    fn correct(&self, val: u8) -> u16 {
        self.lut[val as usize]
    }
}

struct Lightbar {
    gamma_lut: GammaLut,
    mbi: MBI5043<AnyPin, AnyPin, AnyPin>,
}

impl Lightbar {
    fn new(sdi: AnyPin, sck: AnyPin, le: AnyPin, gamma: f32) -> Self {
        Self {
            mbi: MBI5043::new(sdi, sck, le),
            gamma_lut: GammaLut::new(gamma, u16::MAX),
        }
    }
    async fn mbi_send_frame(&self, f: &Frame) {
        for row in 0..HEIGHT as usize {
            pac::GPIOB.outdr().write_value(Outdr(0));
            self.mbi.send_pixels(
                &f.pixels[row * WIDTH as usize..(row + 1) * WIDTH as usize],
                |val| self.gamma_lut.correct(val),
            );
            pac::GPIOB.outdr().write_value(Outdr(1 << row));
            Timer::after_micros(1_000_000 / FPS as u64 / (HEIGHT as u64 + 1)).await;
        }
    }
    async fn display_task(&self) {
        // configure GPIOB as output
        pac::GPIOB
            .cfglr()
            .write_value(gpio::regs::Cfglr(0x33333333));
        pac::GPIOB
            .cfghr()
            .write_value(gpio::regs::Cfghr(0x33333333));

        let mut cmd: Option<MbiCommand> = None;
        let mut draw_frame: Frame = EMPTY_FRAME;

        let mut ticker = Ticker::every(Duration::from_micros(1_000_000 / FPS as u64));
        // TODO: move to interrupt driven
        loop {
            let frame = CUR_FRAME.fetch_add(1, core::sync::atomic::Ordering::Relaxed);
            critical_section::with(|cs| {
                let mut commands = MBI_QUEUE.borrow_ref_mut(cs);

                if commands.peek().is_some_and(|v| frame >= v.when) {
                    cmd = commands.pop();
                } else {
                    cmd = None
                }
            });
            if let Some(cmd) = &cmd {
                match &cmd.inner {
                    MbiCommandInner::Brightness(gain) => self.mbi.set_gain(*gain),
                    MbiCommandInner::Frame(f) => {
                        draw_frame = f.clone();
                    }
                }
            } else {
                self.mbi_send_frame(&draw_frame).await;
                ticker.next().await
            }
        }
    }
}

#[embassy_executor::task]
async fn display_task(app: &'static Lightbar) {
    app.display_task().await
}

#[embassy_executor::main(entry = "qingke_rt::entry")]
// #[qingke_rt::entry]
async fn main(spawner: Spawner) -> ! {
    hal::debug::SDIPrint::enable();
    let mut config = hal::Config::default();
    config.rcc = hal::rcc::Config::SYSCLK_FREQ_144MHZ_HSE;

    let p = hal::init(config);
    Output::new(p.PC13.degrade(), Level::High, Default::default());

    let spi_sck = p.PA5;
    let spi_mosi = p.PA7;
    let le = p.PA4;
    let gclk = p.PA8;

    let ch1 = PwmPin::new_ch1::<0>(gclk);
    let mut pwm = SimplePwm::new(
        p.TIM1,
        Some(ch1),
        None,
        None,
        None,
        Hertz::mhz(8),
        CountingMode::default(),
    );
    let ch = hal::timer::Channel::Ch1;

    pwm.set_duty(ch, 9);
    pwm.enable(ch);

    let app: Lightbar = Lightbar::new(spi_mosi.degrade(), spi_sck.degrade(), le.degrade(), 2.8);

    spawner.spawn(usb(p.USBD, Irqs, p.PA12, p.PA11)).unwrap();
    spawner
        .spawn(blink(p.PC14.degrade(), p.PC15.degrade()))
        .unwrap();

    loop {
        app.display_task().await
    }
}

#[inline(never)]
#[panic_handler]
fn panic(_info: &PanicInfo) -> ! {
    loop {
        println!("{:?}", _info);
        core::sync::atomic::compiler_fence(core::sync::atomic::Ordering::SeqCst);
    }
}
